﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;
using System.ComponentModel;
using System.IO;

using CamBam;
using CamBam.Geom;
using CamBam.CAD;
using CamBam.UI;
using CamBam.Util;

namespace ExcellonIO
{
    /// <summary>
    /// This class is used to store and retreive configuration settings used by the plugin.
    /// </summary>
    public class ExcellonIOConfig 
    {
        public static ExcellonIOConfig Defaults;

        [Category("Excellon")]
        [DefaultValue(1e-4)]
        [Description("The scaling factor is applied only to values with no decimal point.")]
        [DisplayName("X Scale (Inches)")]
        public double XScale { get; set; }

        [Category("Excellon")]
        [DefaultValue(1e-4)]
        [Description("The scaling factor is applied only to values with no decimal point.")]
        [DisplayName("Y Scale (Inches)")]
        public double YScale { get; set; }

        [Category("Excellon")]
        [DefaultValue(1e-2)]
        [Description("The scaling factor is applied only to values with no decimal point.")]
        [DisplayName("X Scale (Metric)")]
        public double XScaleMetric { get; set; }

        [Category("Excellon")]
        [DefaultValue(1e-2)]
        [Description("The scaling factor is applied only to values with no decimal point.")]
        [DisplayName("Y Scale (Metric)")]
        public double YScaleMetric { get; set; }


        [Category("Excellon")]
        [DefaultValue(true)]
        [Description("Create drilling machine operation.")]
        [DisplayName("Create MOP")]
        public bool CreateMOP { get; set; }

        [Category("Excellon")]
        [DefaultValue(-0.1)]
        [Description("Target depth for drilling operations.")]
        [DisplayName("Target Depth")]
        public double TargetDepth { get; set; }

        [Category("Excellon")]
        [DefaultValue(true)]
        [Description("Set the tool diameter in the drilling machine operation.")]
        [DisplayName("Set Tool Diameter")]
        public bool SetToolDiameter { get; set; }

        [Category("Excellon")]
        [DefaultValue(false)]
        [Description("Set the hole diameter in the drilling machine operation.")]
        [DisplayName("Set Hole Diameter")]
        public bool SetHoleDiameter { get; set; }

        [Category("Excellon")]
        [DefaultValue(false)]
        [Description("Set the tool profile in the drilling machine operation.")]
        [DisplayName("Set Tool Profile")]
        public bool SetToolProfile { get; set; }

        [Category("Excellon")]
        [DefaultValue(typeof(CamBam.CAM.ToolProfiles),"Drill")]
        [Description("Set the tool profile in the drilling machine operation.")]
        [DisplayName("Tool Profile")]
        public CamBam.CAM.ToolProfiles ToolProfile { get; set; }
                
        public ExcellonIOConfig()
        {
            XScale = 1e-4;
            YScale = 1e-4;
            XScaleMetric = 1e-2;
            YScaleMetric = 1e-2;
            CreateMOP = true;
            TargetDepth = -0.1;
            SetToolDiameter = true;
            SetHoleDiameter = false;
            SetToolProfile = false;
            ToolProfile = CamBam.CAM.ToolProfiles.Drill;
        }

        public static string UserConfigFilename
        {
            get
            {
                return CamBam.FileUtils.GetFullPath(CamBamConfig.Defaults.SystemPath, "ExcellonIO.config");
            }
        }

        public static bool WriteConfigFile()
        {
            if (Defaults == null) return false;
            
            try
            {
                FileInfo fi = new FileInfo(UserConfigFilename);
                if (!fi.Directory.Exists)
                {
                    fi.Directory.Create();
                }

                XmlSerializer serializer = new XmlSerializer(Defaults.GetType());

                using (TextWriter writer = new StreamWriter(UserConfigFilename, false))
                {
                    serializer.Serialize(writer, Defaults);
                    writer.Close();
                }
            }
            catch (Exception e)
            {
                ThisApplication.HandleException(e);
                return false;
            }
            return true;
        }

        public static void ReadConfigFile()
        {
            ExcellonIOConfig cfg = null;

            try
            {
                if (File.Exists(UserConfigFilename))
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(ExcellonIOConfig));

                    using (TextReader reader = new StreamReader(UserConfigFilename))
                    {
                        cfg = (ExcellonIOConfig)(serializer.Deserialize(reader));

                        reader.Close();
                    }
                }
            }
            catch (Exception e)
            {
                ThisApplication.MsgBox(String.Format("Error reading config file : {0}\r\n{1}", UserConfigFilename, e.Message));
            }

            if (cfg == null)
                cfg = new ExcellonIOConfig();

            Defaults = cfg;
        }
    }
}
