﻿/* 
 * Excellon programming reference...
 * http://www.excellon.com/manuals/program.htm
 * 
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;

using CamBam;
using CamBam.CAD;
using CamBam.CAM;
using CamBam.Geom;
using CamBam.Util;
using CamBam.Values;

namespace ExcellonIO
{
    /// <summary>
    /// This is a file handler class for Excellon files
    /// 
    /// The class should handle the following overrides...
    /// 
    /// public string FileFilter {get;}
    ///     A property that returns the file open filter and file extensions that this class handles.
    ///     
    /// public bool ReadFile(string path)
    ///     Method called by CamBam to read the source file
    ///     
    /// public bool AddToDocument(CADFile doc)
    ///     If a file is read successfully, CamBam will call this method to add the file contents to a new drawing.
    /// 
    /// </summary>
    public class ExcellonIOHandler : CADFileIO
    {
        public bool InHeader = false;
        // assume inches
        public Units DrawingInits = Units.Inches;
        public bool LeadingZeros = true;
        protected List<Entity> _Points;
        protected List<int> _HoleSizes;
        protected PointList _CurrentPoints;
        public Dictionary<int, double> Tools;
        protected double _xscale = 1e-4;
        protected double _yscale = 1e-4;
        protected Point2F _CurrentPoint = Point2F.Undefined;

        public Point2F CurrentPoint
        {
            get
            {
                return _CurrentPoint;
            }
        }

        /// <summary>
        /// A property that returns the file open filter and file extensions that this class handles.
        /// </summary>
        public override string FileFilter
        {
            get
            {
                return "Excellon File (*.drl)|*.drl";
            }
        }

        /// <summary>
        /// Method called by CamBam to read the source file
        /// </summary>
        /// <param name="path">The path and filename of the file to read</param>
        /// <returns>True if the file was read successfully.</returns>
        public override bool ReadFile(string path)
        {
            _Initialize();

            using (StreamReader sr = new StreamReader(path))
            {
                long line_num = 0;

                // makes sure there is a current point list...
                ChangeTool(0);

                for (; ; )
                {
                    line_num++;
                    string s = sr.ReadLine();
                    if (s == null) break;

                    s = s.Trim();
                    if (s.Length == 0) continue;

                    if (s == "%")
                    {
                        InHeader = false;
                        continue;
                    }

                    // Find and process the first pattern that matches this string...
                    foreach( _ExcellonPattern pattern in _Patterns )
                    {
                        if (pattern.Process(this, s)) break;
                    }
                }
                ThisApplication.AddLogMessage("{0} lines read from {1}", line_num, path);
            }

            return true;
        }

        /// <summary>
        /// If a file is read successfully, CamBam will call this method to add the file contents to a new drawing.
        /// </summary>
        /// <param name="doc">The CADFile to add the source file contents to.</param>
        /// <returns>True if the document was updated successfully.</returns>
        public override bool AddToDocument(CADFile doc)
        {
            doc.DrawingUnits = DrawingInits;

            doc.Add(_Points);

            if (ExcellonIOConfig.Defaults.CreateMOP)
            {
                // Make sure there is an active part in the drawing...
                doc.EnsureActivePart(false);

                for (int i = 0; i < _Points.Count; i++)
                {
                    PointList plist = (PointList)_Points[i];

                    MOPDrill drill = new MOPDrill();
                    drill.PrimitiveIds = new int[] { plist.ID };
                    drill.Name = drill.GetSafeName(doc.ActivePart);
                    int tool_num = _HoleSizes[i];
                    double tool_diameter = 0;
                    if (Tools.ContainsKey(tool_num))
                    {
                        tool_diameter = Tools[tool_num];
                    }

                    drill.ToolNumber = new CBValue<int>(tool_num);
                    
                    if (ExcellonIOConfig.Defaults.SetToolDiameter)
                        drill.ToolDiameter = new CBValue<double>(tool_diameter);
                    
                    drill.TargetDepth = new CBValue<double>(ExcellonIOConfig.Defaults.TargetDepth);
                    
                    if (ExcellonIOConfig.Defaults.SetToolProfile)
                        drill.ToolProfile = new CBValue<ToolProfiles>(ExcellonIOConfig.Defaults.ToolProfile);
                    
                    if (ExcellonIOConfig.Defaults.SetHoleDiameter)
                        drill.HoleDiameter = new CBValue<double>(tool_diameter);
                    
                    doc.ActivePart.MachineOps.Add(drill);
                }
            }

            return true;
        }

        int _CurrentTool = 0;

        /// <summary>
        /// Handle a tool number change (T# command)
        /// </summary>
        /// <param name="newtool">Tool number</param>
        protected void ChangeTool(int newtool)
        {
            if (_CurrentTool == newtool) return;

            // Start a new point list
            if (_CurrentPoints == null || _CurrentPoints.Points.Count > 0)
            {
                _CurrentPoints = new PointList();
                _Points.Add(_CurrentPoints);
                _HoleSizes.Add(newtool);
            }
            else
            {
                // The current point list is empty, so just update it's hole size
                _HoleSizes[_Points.Count - 1] = newtool;
            }

            if (Tools.ContainsKey(newtool))
            {
                double td = Tools[newtool];
                _CurrentPoints.Tag = td.ToString();
            }
            else
            {
                // WARNING! Tool not defined.
                ThisApplication.AddLogMessage("WARNING! Tool {0} not defined.", newtool);
                _CurrentPoints.Tag = "0";
            }
        }

        /// <summary>
        /// Add a new drill point to the current point list
        /// </summary>
        /// <param name="p">drill point to add</param>
        protected void AddPoint(Point2F p)
        {
            _CurrentPoint = p;
            _CurrentPoints.Add(_CurrentPoint);
        }

        /// <summary>
        /// Add a new drill point to the current point list
        /// </summary>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        protected void AddPoint(double x, double y)
        {
            _CurrentPoint = new Point2F(x, y);
            _CurrentPoints.Add(_CurrentPoint);
        }

        // A list of patterns that are searched to find matches with current file lines
        static List<_ExcellonPattern> _Patterns = null;

        /// <summary>
        /// Make sure all instance and static variables are defined.
        /// </summary>
        protected void _Initialize()
        {
            // assume inches
            DrawingInits = Units.Inches;
            LeadingZeros = true;
            
            _Points = new List<Entity>();
            _HoleSizes = new List<int>();
            Tools = new Dictionary<int, double>();

            _CurrentPoints = null;
            _CurrentPoint = Point2F.Undefined;

            if (ExcellonIOConfig.Defaults == null)
            {
                ExcellonIOConfig.ReadConfigFile();
            }

            _xscale = ExcellonIOConfig.Defaults.XScale;
            _yscale = ExcellonIOConfig.Defaults.YScale;

            if (_Patterns != null) return;

            // Add the patters to the patter list.
            // The patterns will be processed in the order they appear in this list, to find the first match...

            _Patterns = new List<_ExcellonPattern>() { 
                new _ExcellonPattern_MCode(),
                new _ExcellonPattern_Units(),
                new _ExcellonPattern_GCode(),
                new _ExcellonPattern_ToolDef(),
                new _ExcellonPattern_SetTool(),
                new _ExcellonPattern_PointXY(),
                new _ExcellonPattern_PointXY_NoDP()
            };
        }

        #region Excellon Patterns

        /// <summary>
        /// A pattern is just a regular expression, with a method that searches a string,
        /// and process it if a match is found.
        /// </summary>
        class _ExcellonPattern
        {
            public Regex Pattern;

            protected _ExcellonPattern(string sregex)
            {
                Pattern = new Regex(sregex, RegexOptions.Compiled | RegexOptions.IgnoreCase);
            }

            /// <summary>
            /// Each pattern should overide this method
            /// </summary>
            /// <param name="io">The current file handler instance</param>
            /// <param name="str">The file line being read</param>
            /// <returns>True if the pattern processes this string</returns>
            public virtual bool Process(ExcellonIOHandler io, string str)
            {
                Match m = Pattern.Match(str);
                if (!m.Success) return false;
                return true;
            }
        }

        /// <summary>
        /// G-Code pattern   G##
        /// </summary>
        class _ExcellonPattern_GCode : _ExcellonPattern
        {
            public _ExcellonPattern_GCode()
                : base(@"^G([0-9]+)")
            {
            }

            public override bool Process(ExcellonIOHandler io, string str)
            {
                Match m = Pattern.Match(str);
                if (!m.Success) return false;

                int gcode = int.Parse(m.Groups[1].ToString());

                switch (gcode)
                {
                    case 90:
                        // absolute mode
                        break;
                }

                return true;
            }
        }

        /// <summary>
        /// M-Ccode pattern   M##
        /// </summary>
        class _ExcellonPattern_MCode : _ExcellonPattern
        {
            public _ExcellonPattern_MCode()
                : base(@"^M([0-9]+)")
            {
            }

            public override bool Process(ExcellonIOHandler io, string str)
            {
                Match m = Pattern.Match(str);
                if (!m.Success) return false;

                int gcode = int.Parse(m.Groups[1].ToString());

                switch (gcode)
                {
                    case 0: // X#Y#
                        // end of program, no rewind
                        break;
                    case 1:
                        // end pattern
                        break;
                    case 2: // X#Y#
                        // repeat pattern offset 
                        break;
                    case 6: // X#Y#
                        // optional stop
                        break;
                    case 8:
                        // end of step and repeat
                        break;
                    case 9:
                        // stop for inspection
                        break;
                    case 14:
                        // Z axis route position with depth controlled contouring
                        break;
                    case 15:
                        // Z axis route position 
                        break;
                    case 16:
                        // retract with clamping
                        break;
                    case 17:
                        // retract without clamping
                        break;
                    case 18:
                        // command tool tip check
                        break;
                    case 25: 
                        // beginning of pattern
                        break;
                    case 30: // X#Y#
                        // end of program rewind
                        break;
                    case 45: // ,log message\
                        // long operator message on multiple\ part program lines 
                        break;
                    case 47: // ,text
                        // operator message
                        break;
                    case 48:
                        // start of header
                        io.InHeader = true;
                        break;
                    case 71:
                        // metric measuring mode
                        io.DrawingInits = Units.Millimeters;
                        io._xscale = ExcellonIOConfig.Defaults.XScaleMetric;
                        io._yscale = ExcellonIOConfig.Defaults.YScaleMetric;
                        break;
                    case 72:
                        // inch measuring mode
                        io.DrawingInits = Units.Inches;
                        io._xscale = ExcellonIOConfig.Defaults.XScale;
                        io._yscale = ExcellonIOConfig.Defaults.YScale;
                        break;
                }

                return true;
            }
        }

        /// <summary>
        /// Units pattern   INCH/METRIC(,LZ)
        /// </summary>
        class _ExcellonPattern_Units : _ExcellonPattern
        {
            public _ExcellonPattern_Units()
                : base(@"^(INCH|METRIC)\W?(LZ|TZ)?")
            {
            }

            public override bool Process(ExcellonIOHandler io, string str)
            {
                Match m = Pattern.Match(str);
                if (!m.Success) return false;

                string units = m.Groups[1].ToString().ToUpper();

                if (units == "METRIC")
                {
                    io.DrawingInits = Units.Millimeters;
                    io._xscale = ExcellonIOConfig.Defaults.XScaleMetric;
                    io._yscale = ExcellonIOConfig.Defaults.YScaleMetric;
                }
                else
                {
                    io.DrawingInits = Units.Inches;
                    io._xscale = ExcellonIOConfig.Defaults.XScale;
                    io._yscale = ExcellonIOConfig.Defaults.YScale;
                }

                if (m.Groups[2].Length > 0)
                {
                    string zz = m.Groups[2].ToString().ToUpper();
                    if (zz == "LZ")
                        io.LeadingZeros = true;
                    else if (zz == "TZ")
                        io.LeadingZeros = false;
                }
                return true;
            }
        }

        /// <summary>
        /// Tool definition pattern   T##C####
        /// </summary>
        class _ExcellonPattern_ToolDef : _ExcellonPattern
        {
            public _ExcellonPattern_ToolDef()
                : base(@"^T([0-9]+)(C)([0-9]*\.?[0-9]*)")
            {
            }

            public override bool Process(ExcellonIOHandler io, string str)
            {
                if (!io.InHeader) return false;
                Match m = Pattern.Match(str);
                if (!m.Success) return false;

                int tool_number = int.Parse(m.Groups[1].ToString());


                double diameter = double.Parse(m.Groups[3].ToString());

                if (io.Tools.ContainsKey(tool_number))
                {
                    io.Tools[tool_number] = diameter;
                }
                else
                {
                    io.Tools.Add(tool_number, diameter);
                }
                return true;
            }
        }

        /// <summary>
        /// Set tool pattern   T##
        /// </summary>
        class _ExcellonPattern_SetTool : _ExcellonPattern
        {
            public _ExcellonPattern_SetTool()
                : base(@"^T([0-9]+)")
            {
            }

            public override bool Process(ExcellonIOHandler io, string str)
            {
                if (io.InHeader) return false;
                Match m = Pattern.Match(str);
                if (!m.Success) return false;

                int tool_number = int.Parse(m.Groups[1].ToString());

                io.ChangeTool(tool_number);

                return true;
            }
        }

        /// <summary>
        /// Point pattern   X####Y#####
        /// (without decimal points)
        /// </summary>
        class _ExcellonPattern_PointXY_NoDP : _ExcellonPattern
        {
            public _ExcellonPattern_PointXY_NoDP()
                : base(@"^(?:X([+-]?[0-9]+))?(?:Y([+-]?[0-9]+))?")
            {
            }

            public override bool Process(ExcellonIOHandler io, string str)
            {
                if (io.InHeader) return false;
                Match m = Pattern.Match(str);
                if (!m.Success) return false;

                if (string.IsNullOrEmpty(m.Groups[1].ToString()) && string.IsNullOrEmpty(m.Groups[2].ToString())) return false;

                Point2F pos = io.CurrentPoint;

                string px = m.Groups[1].ToString();
                if (!string.IsNullOrEmpty(px))
                {
                    if (io.LeadingZeros && px.Length < 6)
                        px += new string('0', 6 - px.Length);
                    pos.X = double.Parse(px) * io._xscale;
                }

                string py = m.Groups[2].ToString();
                if (!string.IsNullOrEmpty(py))
                {
                    if (io.LeadingZeros && py.Length < 6)
                        py += new string('0', 6 - py.Length);
                    pos.Y = double.Parse(py) * io._yscale;
                }

                io.AddPoint(pos);

                return true;
            }
        }

        /// <summary>
        /// Point pattern   X##.###Y##.###
        /// (with decimal points)
        /// </summary>
        class _ExcellonPattern_PointXY : _ExcellonPattern
        {
            public _ExcellonPattern_PointXY()
                : base(@"^(?:X([+-]{0,1}[0-9]*\.{1}[0-9]*))?(?:Y([+-]{0,1}[0-9]*\.{1}[0-9]*))?")
            {
            }

            public override bool Process(ExcellonIOHandler io, string str)
            {
                if (io.InHeader) return false;
                Match m = Pattern.Match(str);
                if (!m.Success) return false;

                if (string.IsNullOrEmpty(m.Groups[1].ToString()) && string.IsNullOrEmpty(m.Groups[2].ToString())) return false;

                Point2F pos = io.CurrentPoint;

                if (!string.IsNullOrEmpty(m.Groups[1].ToString()))
                    pos.X = double.Parse(m.Groups[1].ToString());
                if (!string.IsNullOrEmpty(m.Groups[2].ToString()))
                    pos.Y = double.Parse(m.Groups[2].ToString());


                io.AddPoint(pos);

                return true;
            }
        }

        #endregion
    }
}
